from tracardi.domain.event import Event
from tracardi.service.storage.factory import StorageFor


async def load_event_by_type(event_type):
    return await StorageFor.crud('event', class_type=Event).load_by('type', event_type, limit=1)



from typing import List
from tracardi.domain.entity import Entity
from tracardi.config import elastic, tracardi
from tracardi.domain.profile import Profile
from tracardi.service.storage.factory import StorageFor, storage_manager
from tracardi.service.storage.profile_cacher import ProfileCache


async def load_merged_profile(id: str) -> Profile:
    """
    Loads current profile. If profile was merged then it loads merged profile.
    """

    if tracardi.cache_profiles is not False:
        profile_cache = ProfileCache()
        if profile_cache.exists(id):
            return profile_cache.get_profile(id)

    entity = Entity(id=id)
    profile = await StorageFor(entity).index('profile').load(Profile)  # type: Profile
    if profile is not None and profile.mergedWith is not None:
        # Has merged profile
        profile = await load_merged_profile(profile.mergedWith)

    return profile


async def load_profiles_to_merge(merge_key_values: List[tuple], limit=1000) -> List[Profile]:
    profiles = await storage_manager('profile').load_by_values(merge_key_values, limit)
    return [Profile(**profile) for profile in profiles]


async def save_profile(profile: Profile):
    if tracardi.cache_profiles is not False:
        cache = ProfileCache()
        cache.save_profile(profile)
    result = await StorageFor(profile).index().save()
    if elastic.refresh_profiles_after_save:
        await storage_manager('profile').flush()
    return result


class TracardiException(Exception):
    pass


class StorageException(TracardiException):
    pass


class ExpiredException(TracardiException):
    pass


class UnauthorizedException(TracardiException):
    pass





from tracardi.service.storage.factory import storage_manager


async def load_by_profile(profile_id: str):
    return await storage_manager('console-log').load_by("profile_id", profile_id)


async def load_by_event(event_id: str):
    return await storage_manager('console-log').load_by("event_id", event_id)



import asyncio

from tracardi.service.singleton import Singleton
from tracardi_dot_notation.dot_accessor import DotAccessor

from tracardi.process_engine.tql.parser import Parser
from tracardi.process_engine.tql.transformer.expr_transformer import ExprTransformer


class Condition(metaclass=Singleton):

    def __init__(self):
        self.parser = Parser(Parser.read('grammar/uql_expr.lark'), start='expr')

    def parse(self, condition):
        return self.parser.parse(condition)

    async def evaluate(self, condition, dot: DotAccessor):
        # todo cache tree
        tree = self.parse(condition)
        await asyncio.sleep(0)
        return ExprTransformer(dot=dot).transform(tree)




import asyncio

from tracardi.service.storage.driver import storage
from tracardi.service.storage.factory import StorageFor
from tracardi_plugin_sdk.domain.register import Plugin, Spec, MetaData
from tracardi_plugin_sdk.domain.result import Result
from tracardi_plugin_sdk.action_runner import ActionRunner

from tracardi.domain.entity import Entity
from tracardi.domain.event import Event
from tracardi.domain.profile import Profile
from tracardi.domain.session import Session


class DebugPayloadAction(ActionRunner):

    def __init__(self, **kwargs):
        if 'event' not in kwargs or 'type' not in kwargs['event']:
            raise ValueError("Please define event.type in config section.")
        self.event_type = kwargs['event']['type']

    async def run(self, **kwargs):
        if self.debug:
            result = await storage.driver.event.load_event_by_type(self.event_type)

            if result.total == 0:
                raise ValueError(
                    "There is no event with type `{}`. Check configuration for correct event type.".format(
                        self.event_type))

            event_data = list(result)[0]

            self.event = Event(**event_data)

            profile_entity = Entity(id=self.event.profile.id)
            session_entity = Entity(id=self.event.session.id)
            profile_task = asyncio.create_task(StorageFor(profile_entity).index('profile').load(Profile))
            session_task = asyncio.create_task(StorageFor(session_entity).index('session').load(Session))

            profile = await profile_task
            session = await session_task

            if session is None:
                raise ValueError(
                    "Event id `{}` has reference to empty session id `{}`. Debug stopped. This event is corrupted.".format(
                        self.event.id, self.event.session.id))

            if profile is None:
                raise ValueError(
                    "Event type `{}` has reference to empty profile id `{}`. Debug stopped. This event is corrupted.".format(
                        self.event.id, self.event.profile.id))

            self.profile.replace(profile)
            self.session.replace(session)

            return Result(port="event", value=self.event.dict())

        return Result(port="event", value=self.event.dict())


def register() -> Plugin:
    return Plugin(
        start=True,
        debug=True,
        spec=Spec(
            module='tracardi.process_engine.action.v1.debug_payload_action',
            className='DebugPayloadAction',
            inputs=[],
            outputs=["event"],
            init={
                "event": {
                    "type": None,
                }
            },
            manual="debug_payload_action",
            version='0.1',
            license="MIT",
            author="Risto Kowaczewski"
        ),
        metadata=MetaData(
            name='Debug',
            desc='Loads debug payload into flow. This action is executed only in debug mode. ' +
                 'Use it to inject payload defined it config to analyse you workflow.',
            keywords=['start node'],
            type='flowNode',
            width=100,
            height=100,
            icon='debug',
            group=["Input/Output"]
        )
    )



import aioredis
import redis
from tracardi.service.singleton import Singleton


class AsyncRedisClient(metaclass=Singleton):
    def __init__(self, host):
        self.client = aioredis.from_url(host)


class RedisClient(metaclass=Singleton):
    def __init__(self, host):
        self.client = redis.from_url(host)



import json

from tracardi.config import redis_config

from tracardi.domain.profile import Profile
from tracardi.service.singleton import Singleton
from tracardi.service.storage.redis_client import RedisClient


class ProfileCache(metaclass=Singleton):

    def __init__(self):
        self.redis = RedisClient(redis_config.redis_host)
        self.hash = "profile-cache"

    def exists(self, id):
        return self.redis.client.hexists(self.hash, id)

    def get_profile(self, id) -> Profile:
        profile_bson = self.redis.client.hget(self.hash, id)
        return Profile(**json.loads(profile_bson.decode('utf-8')))

    def save_profile(self, profile: Profile):
        return self.redis.client.hset(self.hash, profile.id, profile.json())



from tracardi.service.storage.factory import storage_manager


async def refresh():
    return await storage_manager('profile').refresh()


async def flush():
    return await storage_manager('profile').flush()



from tracardi.domain.entity import Entity
from tracardi.domain.value_object.storage_info import StorageInfo


class Project(Entity):
    name: str

    # def storage(self) -> StorageCrud:
    #     return StorageCrud("project", Project, entity=self)

    @staticmethod
    def storage_info() -> StorageInfo:
        return StorageInfo(
            'project',
            Project
        )



from datetime import datetime
from typing import Optional, List, Tuple, Any
from pydantic import BaseModel
from tracardi.domain.event import Event

from ..entity import Entity
from ..payload.event_payload import EventPayload
from ..profile import Profile
from ..session import Session
from ..metadata import Metadata
from ..time import Time


class TrackerPayload(BaseModel):
    source: Entity
    session: Entity

    metadata: Optional[Metadata]
    profile: Optional[Entity] = None
    context: Optional[dict] = {}
    properties: Optional[dict] = {}
    events: List[EventPayload] = []
    options: Optional[dict] = {}

    def __init__(self, **data: Any):
        data['metadata'] = Metadata(
            time=Time(
                insert=datetime.utcnow()
            ))
        super().__init__(**data)

    def get_events(self, session: Session, profile: Profile) -> List[Event]:
        _events = []
        if self.events:
            for event in self.events:  # type: EventPayload
                _event = event.to_event(self.metadata, self.source, session, profile, event.options)
                _events.append(_event)
        return _events

    def return_profile(self):
        return self.options and "profile" in self.options and self.options['profile'] is True

    def is_disabled(self, key):
        return key in self.options and self.options[key] is False

    async def get_profile_and_session(self, session: Session, load_merged_profile) -> Tuple[Profile, Session]:

        """
        Returns session. Creates profile if it does not exist.If it exists connects session with profile.
        """

        is_new_profile = False
        is_new_session = False

        if session is None:  # loaded session is empty

            session = Session(id=self.session.id)
            is_new_session = True

            # Bind profile

            if self.profile is None:

                # Create empty default profile generate profile_id
                profile = Profile.new()

                # Create new profile
                is_new_profile = True

            else:

                # Id exists load profile from storage
                profile = await load_merged_profile(id=self.profile.id)  # type: Profile

                if profile is None:
                    # Profile id delivered but profile does not exist in storage.
                    # Id was forged

                    profile = Profile.new()

                    # Create new profile
                    is_new_profile = True

        else:

            # There is session. Copy profile id form session to profile

            if session.profile is not None:
                # Loaded session has profile

                # Load profile based on profile id saved in session
                profile = await load_merged_profile(id=session.profile.id)  # type: Profile

                if session.profile.id != profile.id:
                    # Profile in session id has been merged. Change profile in session.

                    session.profile.id = profile.id
                    session.metadata.time = Time(insert=datetime.utcnow())

                    is_new_session = True

                # profile = await profile.storage().load()  # type: Profile

            else:
                # Corrupted session

                profile = None

            if profile is None:
                # Id exists but profile not exist in storage.

                profile = Profile.new()

                # Create new profile
                is_new_profile = True

        session.context = self.context
        session.properties = self.properties
        session.operation.new = is_new_session

        profile.operation.new = is_new_profile

        return profile, session



from tracardi.domain.resource import Resource
from tracardi.service.storage.driver import storage


async def read_source(resource_id: str) -> Resource:
    return await storage.driver.resource.load(resource_id)



from datetime import datetime
import os


class Index:
    def __init__(self, multi_index, index, mapping, rel):
        self.multi_index = multi_index
        self.rel = rel
        self.index = index
        self.prefix = "{}-".format(os.environ['INSTANCE_PREFIX']) if 'INSTANCE_PREFIX' in os.environ else ''
        self.mapping = mapping

    def _index(self):
        return self.prefix + self.index

    def get_read_index(self):
        if self.multi_index is False:
            return self._index()

        return self._index() + "*"

    def get_write_index(self):
        if self.multi_index is False:
            return self._index()

        date = datetime.now()
        return self._index() + f"-{date.year}-{date.month}"


class Resource:

    def __init__(self):
        self.resources = {
            "project": Index(multi_index=False, index="tracardi-flow-project", mapping=None, rel=None),
            "action": Index(multi_index=False, index="tracardi-flow-action-plugins", mapping=None, rel=None),
            "token": Index(multi_index=False, index="tracardi-token", mapping="mappings/token-index.json", rel=None),
            "resource": Index(multi_index=False, index="tracardi-resource", mapping=None, rel=None),
            "session": Index(multi_index=True, index="tracardi-session", mapping=None, rel='profile.id'),
            "profile": Index(multi_index=False, index="tracardi-profile", mapping="mappings/profile-index.json",
                             rel='_id'),
            "event": Index(multi_index=True, index="tracardi-event", mapping="mappings/event-index.json", rel=None),
            "flow": Index(multi_index=False, index="tracardi-flow", mapping=None, rel=None),
            "rule": Index(multi_index=False, index="tracardi-rule", mapping=None, rel=None),
            "segment": Index(multi_index=False, index="tracardi-segment", mapping="mappings/segment-index.json",
                             rel=None),
            "console-log": Index(multi_index=False, index="tracardi-console-log",
                                 mapping="mappings/console-log-index.json", rel=None),
            "stat-log": Index(multi_index=False, index="tracardi-stat-log", mapping=None, rel=None),
            "debug-info": Index(multi_index=False, index="tracardi-debug-info",
                                mapping="mappings/debug-info-index.json", rel=None),
            "api-instance": Index(multi_index=False, index="tracardi-api-instance",
                                  mapping="mappings/api-instance-index.json", rel=None),
            "task": Index(multi_index=False, index="tracardi-task", mapping="mappings/task-index.json", rel=None)
        }

    def __getitem__(self, item):
        return self.resources[item]

    def __contains__(self, item):
        return item in self.resources


resources = Resource()



# from tracardi.service.storage.collection_crud import CollectionCrud
#
#
# class Resources(list):
#
#     # Persistence
#     def bulk(self) -> CollectionCrud:
#         return CollectionCrud("resource", self)



from typing import List, Tuple

from tracardi.domain.resource import Resource, ResourceRecord

from tracardi.service.storage.factory import storage_manager, StorageForBulk


async def refresh():
    return await storage_manager('resource').refresh()


async def flush():
    return await storage_manager('resource').flush()


async def load_all() -> Tuple[List[Resource], int]:
    result = await StorageForBulk().index('resource').load()
    data = [ResourceRecord.construct(Resource.__fields_set__, **r).decode() for r in result]
    return data, result.total



from datetime import datetime
from typing import Optional, Any
from .metadata import Metadata
from .named_entity import NamedEntity
from .time import Time
from .type import Type
from .value_object.storage_info import StorageInfo


class Rule(NamedEntity):
    event: Type
    flow: NamedEntity

    source: Optional[NamedEntity] = None
    enabled: Optional[bool] = True
    description: Optional[str] = "No description provided"
    properties: Optional[dict] = None
    metadata: Optional[Metadata]

    def __init__(self, **data: Any):
        data['metadata'] = Metadata(
            time=Time(
                insert=datetime.utcnow()
            ))
        super().__init__(**data)

    @staticmethod
    def storage_info() -> StorageInfo:
        return StorageInfo(
            'rule',
            Rule
        )



from tracardi.domain.record.flow_action_plugin_record import FlowActionPluginRecord

from tracardi.domain.flow_action_plugin import FlowActionPlugin
from tracardi.service.storage.factory import StorageFor


async def save_plugin(action_id, plugin_data):
    action_plugin = FlowActionPlugin(id=action_id, plugin=plugin_data)
    record = FlowActionPluginRecord.encode(action_plugin)
    return await StorageFor(record).index().save()



from datetime import datetime
from typing import Optional, Any
from .entity import Entity
from .metadata import Metadata
from .time import Time
from .value_object.storage_info import StorageInfo
from ..service.secrets import encrypt, decrypt


class Resource(Entity):
    type: str
    metadata: Optional[Metadata]
    name: Optional[str] = "No name provided"
    description: Optional[str] = "No description provided"
    config: Optional[dict] = {}
    origin: str = "event"
    enabled: Optional[bool] = True
    consent: bool = False

    def __init__(self, **data: Any):
        data['metadata'] = Metadata(
            time=Time(
                insert=datetime.utcnow()
            ))
        super().__init__(**data)

    # # Persistence
    # def storage(self) -> StorageCrud:
    #     return StorageCrud("resource", Resource, entity=self)

    @staticmethod
    def storage_info() -> StorageInfo:
        return StorageInfo(
            'resource',
            Resource
        )


class ResourceRecord(Entity):
    type: str
    metadata: Optional[Metadata]
    name: Optional[str] = "No name provided"
    description: Optional[str] = "No description provided"
    config: Optional[str] = None
    enabled: Optional[bool] = True
    origin: str = "event"
    consent: bool = False

    def __init__(self, **data: Any):
        data['metadata'] = Metadata(
            time=Time(
                insert=datetime.utcnow()
            ))
        super().__init__(**data)

    @staticmethod
    def encode(resource: Resource) -> 'ResourceRecord':
        return ResourceRecord(
            id=resource.id,
            name=resource.name,
            description=resource.description,
            type=resource.type,
            origin=resource.origin,
            enabled=resource.enabled,
            consent=resource.consent,
            config=encrypt(resource.config)
        )

    def decode(self) -> Resource:
        return Resource(
            id=self.id,
            name=self.name,
            description=self.description,
            type=self.type,
            origin=self.origin,
            enabled=self.enabled,
            consent=self.consent,
            config=decrypt(self.config)
        )

    # Persistence
    # def storage(self) -> StorageCrud:
    #     return StorageCrud("resource", Resource, entity=self)

    @staticmethod
    def storage_info() -> StorageInfo:
        return StorageInfo(
            'resource',
            ResourceRecord
        )



from datetime import datetime

import pytz
from tracardi_dot_notation.dot_accessor import DotAccessor
from tracardi_plugin_sdk.domain.register import Plugin, Spec, MetaData
from tracardi_plugin_sdk.action_runner import ActionRunner
from tracardi_plugin_sdk.domain.result import Result


class TodayAction(ActionRunner):

    def __init__(self, **kwargs):
        self.timezone = kwargs['timezone'] if 'timezone' in kwargs else None
        self.week_days = ["monday", "tuesday", "wednesday", "thursday", "friday", "saturday", "sunday"]

    async def run(self, payload):
        dot = DotAccessor(self.profile, self.session, payload, self.event, self.flow)

        if self.timezone is not None:
            time_zone = dot[self.timezone]
            time_zone = pytz.timezone(time_zone)
            today = datetime.now(time_zone).today()
        else:
            today = datetime.today()

        return Result(port="payload", value={
            'utcTime': today.utcnow(),
            "dayOfWeek": self.week_days[today.weekday()],
            "day": today.day,
            "month": today.month,
            "year": today.year,
            "week": today.isoweekday(),
            "hour": today.hour,
            "minute": today.minute,
            "second": today.second,
            "ms": today.microsecond,
            "time": today.time(),
            "timestamp": today.timestamp(),
            "fold": today.fold,
            "iso": today.isoformat()
        })


def register() -> Plugin:
    return Plugin(
        start=False,
        spec=Spec(
            module='tracardi.process_engine.action.v1.time.today_action',
            className='TodayAction',
            inputs=["payload"],
            outputs=["payload"],
            version='0.1.1',
            license="MIT",
            author="Risto Kowaczewski",
            init={"timezone": "session@context.time.tz"}

        ),
        metadata=MetaData(
            name='Today',
            desc='Gets today object, that consists day of week, date and current time.',
            type='flowNode',
            width=100,
            height=100,
            icon='today',
            group=["Time"]
        )
    )



from tracardi.domain.value_object.bulk_insert_result import BulkInsertResult
from tracardi.domain.record.event_debug_record import EventDebugRecord
from tracardi.service.storage.factory import StorageForBulk, storage_manager


async def save_debug_info(debug_info_by_event_type_and_rule_name) -> BulkInsertResult:
    records = []
    for debug_info_record in EventDebugRecord.encode(
            debug_info_by_event_type_and_rule_name):  # type: EventDebugRecord
        records.append(debug_info_record)
    # Save in debug index
    return await StorageForBulk(records).index("debug-info").save(replace_id=False)


async def load_by_event(event_id: str):
    return await storage_manager('debug-info').load_by("id", event_id)



from typing import List
from pydantic import BaseModel


class Operation(BaseModel):
    new: bool = False
    update: bool = False
    segment: bool = False
    merge: List[str] = []

    def needs_segmentation(self):
        return self.segment is True

    def needs_update(self):
        return self.update is True

    def needs_merging(self):
        return isinstance(self.merge, list) and len(self.merge) > 0



from tracardi.service.singleton import Singleton
from tracardi.config import tracardi
from tracardi.service.storage.drivers.elastic_driver import ElasticDriver


class Storage(object, metaclass=Singleton):

    def __init__(self):
        self.types = {
            'elastic': ElasticDriver()
        }
        self._driver = self.types[tracardi.storage_driver]

    @property
    def driver(self):
        return self._driver


storage = Storage()



from tracardi.service.storage.factory import storage_manager


async def refresh():
    return await storage_manager('action').refresh()


async def flush():
    return await storage_manager('action').flush()



from tracardi.exceptions.exception import TracardiException
from tracardi.domain.flow import FlowRecord
from tracardi.domain.entity import Entity
from tracardi.service.storage.factory import StorageFor


async def load_flow(flow_id):
    flow_record_entity = Entity(id=flow_id)
    flow_record = await StorageFor(flow_record_entity).index("flow").load(FlowRecord)  # type: FlowRecord
    if not flow_record:
        raise TracardiException("Could not find flow `{}`".format(flow_id))

    return flow_record.decode()







from datetime import datetime
from typing import Any
from pydantic import BaseModel


class Metadata(BaseModel):
    timestamp: datetime = None

    def __init__(self, **data: Any):
        super().__init__(**data)
        self.timestamp = datetime.utcnow()


class Console(BaseModel):
    metadata: Metadata = Metadata()
    event_id: str = None
    flow_id: str = None
    profile_id: str = None
    origin: str
    class_name: str
    module: str
    type: str
    message: str

    # todo cross field validation



import time

from tracardi.service.storage.factory import storage_manager, StorageForBulk


async def load_tasks():
    now = time.time()
    query = {
        "size": 100,
        "query": {
            "bool": {
                "must": [
                    {
                        "range": {
                            "timestamp": {
                                "lte": now
                            }
                        }
                    },
                    {
                        "term": {
                            "status": {
                                "value": "pending"
                            }
                        }
                    }
                ]
            }
        }
    }
    return await storage_manager('task').filter(query)


async def save_tasks(tasks):
    return await StorageForBulk(tasks).index('task').save()



import tracardi.service.storage.drivers.elastic.profile
import tracardi.service.storage.drivers.elastic.session
import tracardi.service.storage.drivers.elastic.event
import tracardi.service.storage.drivers.elastic.events
import tracardi.service.storage.drivers.elastic.flow
import tracardi.service.storage.drivers.elastic.rules
import tracardi.service.storage.drivers.elastic.segment
import tracardi.service.storage.drivers.elastic.plugin
import tracardi.service.storage.drivers.elastic.debug_info
import tracardi.service.storage.drivers.elastic.resource
import tracardi.service.storage.drivers.elastic.flows
import tracardi.service.storage.drivers.elastic.actions
import tracardi.service.storage.drivers.elastic.resources
import tracardi.service.storage.drivers.elastic.profiles
import tracardi.service.storage.drivers.elastic.console_log
import tracardi.service.storage.drivers.elastic.segments
import tracardi.service.storage.drivers.elastic.task


class ElasticDriver:

    @property
    def profile(self):
        return tracardi.service.storage.drivers.elastic.profile

    @property
    def session(self):
        return tracardi.service.storage.drivers.elastic.session

    @property
    def event(self):
        return tracardi.service.storage.drivers.elastic.event

    @property
    def events(self):
        return tracardi.service.storage.drivers.elastic.events

    @property
    def flow(self):
        return tracardi.service.storage.drivers.elastic.flow

    @property
    def flows(self):
        return tracardi.service.storage.drivers.elastic.flows

    @property
    def rules(self):
        return tracardi.service.storage.drivers.elastic.rules

    @property
    def segment(self):
        return tracardi.service.storage.drivers.elastic.segment

    @property
    def plugin(self):
        return tracardi.service.storage.drivers.elastic.plugin

    @property
    def debug_info(self):
        return tracardi.service.storage.drivers.elastic.debug_info

    @property
    def resource(self):
        return tracardi.service.storage.drivers.elastic.resource

    @property
    def actions(self):
        return tracardi.service.storage.drivers.elastic.actions

    @property
    def resources(self):
        return tracardi.service.storage.drivers.elastic.resources

    @property
    def segments(self):
        return tracardi.service.storage.drivers.elastic.segments

    @property
    def profiles(self):
        return tracardi.service.storage.drivers.elastic.profiles

    @property
    def console_log(self):
        return tracardi.service.storage.drivers.elastic.console_log

    @property
    def task(self):
        return tracardi.service.storage.drivers.elastic.task



from typing import Optional, Union
from tracardi.config import memory_cache
from tracardi.domain.entity import Entity
from tracardi.domain.resource import ResourceRecord
from tracardi.event_server.utils.memory_cache import MemoryCache, CacheItem
from tracardi.service.storage.driver import storage
from tracardi.service.storage.factory import StorageFor


class SourceCacher:

    def __init__(self):
        self._cache = MemoryCache()

    async def validate_source(self, source_id) -> ResourceRecord:
        entity = Entity(id=source_id)

        source = await self.get(entity)  # type: Optional[ResourceRecord]
        if source is None:
            raise ValueError("Invalid source.")

        if not source.enabled:
            raise ValueError("Source disabled.")

        return source

    async def get(self, resource: Entity) -> Optional[Union[ResourceRecord]]:
        if 'resource' in self._cache:
            resource = self._cache['resource'].data
            return resource
        else:
            # Expired
            resource = await storage.driver.resource.load_record(resource.id)  # type: ResourceRecord
            if resource is not None:
                self._cache['resource'] = CacheItem(data=resource, ttl=memory_cache.source_ttl)
                return resource
            return None


source_cache = SourceCacher()



from typing import Union, List

from tracardi.domain.event import Event

from tracardi.domain.value_object.bulk_insert_result import BulkInsertResult
from tracardi.domain.value_object.save_result import SaveResult
from tracardi.service.storage.factory import StorageForBulk


async def save_events(events: List[Event], persist_events: bool = True) -> Union[SaveResult, BulkInsertResult]:
    if persist_events:
        events_to_save = [event for event in events if event.is_persistent()]
        event_result = await StorageForBulk(events_to_save).index('event').save()
        event_result = SaveResult(**event_result.dict())

        # Add event types
        for event in events:
            event_result.types.append(event.type)
    else:
        event_result = BulkInsertResult()

    return event_result



import asyncio
from asyncio import Task
from typing import List, Tuple

from tracardi.domain.rule import Rule

from tracardi.domain.event import Event
from tracardi.event_server.utils.memory_cache import MemoryCache, CacheItem
from tracardi.service.storage.factory import storage_manager


def load_rules(events: List[Event]) -> List[Tuple[Task, Event]]:
    return [(
        asyncio.create_task(load_rule(event.type)),
        event
    ) for event in events]


async def load_rule(event_type: str):
    query = {
        "query": {
            "bool": {
                "must": [
                    {
                        "term": {
                            "event.type.keyword": event_type
                        }
                    },
                    {
                        "match": {
                            "enabled": True
                        }
                    }
                ]
            }
        }
    }

    # todo set MemoryCache ttl from env
    memory_cache = MemoryCache()

    if 'rules' not in memory_cache:
        flows_data = await storage_manager(index="rule").filter(query)
        memory_cache['rules'] = CacheItem(data=flows_data, ttl=1)

    rules = list(memory_cache['rules'].data)

    return rules


async def load_flow_rules(flow_id: str) -> List[Rule]:
    rules_attached_to_flow = await storage_manager('rule').load_by('flow.id.keyword', flow_id)
    return [Rule(**rule) for rule in rules_attached_to_flow]


async def refresh():
    return await storage_manager('rule').refresh()


async def flush():
    return await storage_manager('rule').flush()



from tracardi.domain.resource import ResourceRecord, Resource
from tracardi.domain.entity import Entity

from tracardi.service.storage.factory import StorageFor


async def load_record(id: str) -> ResourceRecord:
    return await StorageFor(Entity(id=id)).index("resource").load(ResourceRecord)


async def load(id: str) -> Resource:
    resource_config_record = await load_record(id)  # type: ResourceRecord
    if resource_config_record is None:
        raise ValueError('Resource id {} does not exist.'.format(id))

    return resource_config_record.decode()


async def delete(id: str):
    return await StorageFor(Entity(id=id)).index("resource").delete()



class StorageResult:
    def __init__(self, result=None):
        if result is None:
            self.total = 0
            self._hits = []
            self.chunk = 0
        else:
            self.total = result['hits']['total']['value']
            self._hits = result['hits']['hits']
            self.chunk = len(self._hits)

    def __repr__(self):
        return "hits {}, total: {}".format(self._hits, self.total)

    def __iter__(self):
        for hit in self._hits:
            row = hit['_source']
            row['id'] = hit['_id']
            yield row

    def dict(self):
        return {
            "total": self.total,
            "result": list(self)
        }

    def __len__(self):
        return self.chunk



from tracardi_dot_notation.dict_traverser import DictTraverser
from tracardi_plugin_sdk.domain.register import Plugin, Spec, MetaData
from tracardi_plugin_sdk.action_runner import ActionRunner
from tracardi_plugin_sdk.domain.result import Result
from tracardi_dot_notation.dot_accessor import DotAccessor


class ReshapePayloadAction(ActionRunner):

    def __init__(self, **kwargs):
        self.mapping_template = kwargs

    async def run(self, payload):

        if not isinstance(payload, dict):
            self.console.warning("Payload is not dict that is why you will not be able to read it. ")

        source = DotAccessor(
            self.profile,
            self.session,
            payload,
            self.event,
            self.flow)

        traverser = DictTraverser(source)
        result = traverser.reshape(reshape_template=self.mapping_template)

        return Result(port="payload", value=result)


def register() -> Plugin:
    return Plugin(
        start=False,
        spec=Spec(
            module='tracardi.process_engine.action.v1.traits.reshape_payload_action',
            className='ReshapePayloadAction',
            inputs=["payload"],
            outputs=['payload'],
            init={},
            manual="reshape_payload_action",
            version='0.1',
            license="MIT",
            author="Risto Kowaczewski"
        ),
        metadata=MetaData(
            name='Reshape payload',
            desc='Creates new payload from provided data. Configuration defines where the data should be copied.',
            type='flowNode',
            width=200,
            height=100,
            icon='copy-property',
            group=["Traits"]
        )
    )




from tracardi.domain.profile import Profile
from tracardi.domain.session import Session
from tracardi.domain.value_object.bulk_insert_result import BulkInsertResult
from tracardi.domain.entity import Entity
from tracardi.service.storage.factory import StorageFor, storage_manager


async def save_session(session: Session, profile: Profile, persist_session: bool = True):
    if persist_session and (session.operation.new or session.profile is None or session.profile.id != profile.id):
        # save only profile Entity
        session.profile = Entity(id=profile.id)
        return await StorageFor(session).index().save()
    else:
        return BulkInsertResult()


async def refresh():
    return await storage_manager('session').refresh()


async def flush():
    return await storage_manager('session').flush()
